setreg <- function(object, annotation, pheno.data,verbose = FALSE,
                   exclude.list=NULL,high.conf=50,LIM=0,cnv.abnormality=c("separate","both"),
                   method=c('COVER','COMPOSITE'),
                   cluster.method=c('none','single','complete','average'),cluster.limit=0.8)			    
  {
    ## .....INPUT: 
    ## ........... annotation = a "dataframe" containing probes annotation with the following
    ##                         elements:
    ##                         Name = Probe name, Chr.num = Chromosome number and
    ##                         Pos = Probes position
    ## .......... pheno.data = a data frame of pheno data containing at least sample IDs
    ##                         (columns = sample), and gender (column = gender)
    ##                         (TRUE=male,FALSE=female) plus optional other covariates.
    ##                          Required if CNV analysis performed on X chromosome
    ## .......... exclude.list = vector of 'bad' samples to be excluded from analysis
    ## .......... high.conf = cut.off pctile for 'high confidence' CNV (0-100 range). Default=50
    ##            SC: what if someone wants to use NO confidence information? Set to high.conf=NA
    ## .......... LIM = lower limit of individuals (for creating CNV region),
    ##                   default=0 individuals (no cut-off)
    ## .......... SC: verbose: print some output during computation? This is already implemented in summary

    annotated <- attr(object,"annotated")
    confidence <- attr(object,"confidence")
    input <- attr(object,"input")

    method <- match.arg(method)
    cluster.method <- match.arg(cluster.method)
    cnv.abnormality <- match.arg(cnv.abnormality)
    
    if(!is.na(high.conf) && !confidence)
      stop("High confidence selection criteria requested but no confidence information provided in the data.")
#           " Re-run ",input," to get confidence values.")
    
    if(cluster.limit > 1 || cluster.limit <  0)
      stop("cluster.limit must be between 0 and 1!")
    
    if(cluster.method!='none' & cnv.abnormality=='separate')
      {
        warning("Clustering is only available when cnv.abnormality==both. Clustering set to \"none\"")
        cluster.method <- 'none'
      }

    ## exclusion list
    if(!is.null(exclude.list))
      {
        excl.list = unique(exclude.list)
        
        ##  exclusion from the list
        object <- subset(object,!sample %in% excl.list)
        
        ##...update pheno data
        ## AS: 150908: add exclude.list restriction
        if(!missing(pheno.data))
          pheno.data = subset(pheno.data, !sample %in% excl.list)
      }
    
    
    ## ............................................  no LOH calls

    ## This must be checked.
    ## This is pennCNV specific.
    if(annotated & tolower(input) == "penncnv")
      {
        for(i in names(object)[grep("\\.gene",names(object))])
          object[[i]] <- ifelse(object[[i]]=='NOT_FOUND', '', object[[i]])

        for(i in names(object)[grep("\\.distance",names(object))])
          object[[i]] <- ifelse(object[[i]]=='NOT_FOUND', NA, object[[i]])
      }
    
    ## X chromosome

    if(any(object$chr == "chrX"))
      {
        ##used.X <- TRUE
        if(missing(pheno.data))
          ## @@@@ SC: Better issue an error??
          warning("If X chromosome is used need to supply a pheno.data object to provide samples' gender. All subjects are considered as FEMALES")

        if(!is.logical(pheno.data$gender))
          stop("Gender column must be a logical (male = TRUE)")
        
        male = pheno.data$gender
	pick = object$sample %in% (pheno.data$sample[male])

        object$cn[pick & object$chr == "chrX"] <- object$cn[pick & object$chr == "chrX"] + 1

        object <- subset(object,object$cn != 2)
      }


    # check if the probes in annotation are sorted by chr and pos
    if(!is.numeric(annotation$Chr.num))
      stop("Chromosome numbers have to be numeric!")    
    if(!is.numeric(annotation$Pos))
      stop("Probe positions have to be numeric!")

    ## Why this funny ordering algorithm?
    ## index = annotation$Chr.num*(max(annotation$Pos,na.rm=T)*10) + annotation$Pos
    index <- order(annotation$Chr.num,annotation$Pos)
    annotation <- annotation[index,]

    ##.......START SUMMING OVER INDIVIDUAL.......................


    chr = annotation$Chr.num
    pos = annotation$Pos
    idx = 1:length(chr)
    name = annotation$Name
    names(idx) = name ## SC: quite slow. Better avoid
    composite = method=='COMPOSITE'
    ## overlap = method=='OVERLAP' ## SC: doesn't make sense & not used at all?!
                
    if(!is.na(high.conf))
      {
        if(verbose)
          cat(paste('High confidence CNV is defined as CNV with conf score >=',
                    quantile(object$conf,probs=high.conf/100),sep=""))
        
        pick = object$conf > quantile(object$conf,probs=high.conf/100)
        object = object[pick,]      # choosing a confidence subset here!!
      }
    
    # if conf score is not provided, all individual regions are kept
    # only COVER method is available when no conf score is provided.

    if(is.na(high.conf) || !confidence)
      {
        object$conf = rep(1,nrow(object))

        composite = FALSE
        method = 'COVER'

        if(verbose)
          warning('No confidence score is provided, only COVER method is available!')
    }               

    
    samples  = unique(object$sample)
    cnv.dup = cnv.del = rep(0, length(chr))

    ##...update pheno data
    if(!missing(pheno.data))
      pheno.data <- subset(pheno.data,sample %in% samples)

    
    if(cnv.abnormality=="separate") {
      pick = object$cn < 2  #deletion

      ## This would allow duplicated start/end SNPs
      start = idx[object$startsnp[pick]]
      end =  idx[object$endsnp[pick]]

      ## This won't
      ## start = which(name %in% object$startsnp[pick]) ## indicator of start
      ## end =  which(name %in% object$endsnp[pick]) ## indicator of stop
      
      
      conf =  object$conf[pick]
      
      n = length(cnv.del)
      m = length(start)
      
      call1 = .Fortran("list2seg", 
        as.integer(n),
        seg = as.double(cnv.del),
        as.integer(m),
        as.integer(start),
        as.integer(end),as.double(conf),as.integer(composite),
        PACKAGE="cnvpack")
      
      CNV.del = -call1$seg
      
      pick = object$cn > 2  #duplication
      start = idx[object$startsnp[pick]]
      end =  idx[object$endsnp[pick]]
      conf =  object$conf[pick]
      
      n = length(cnv.dup)
      m = length(start)
      
      call1 = .Fortran("list2seg", 
        as.integer(n),
        seg = as.double(cnv.dup),
        as.integer(m),
        as.integer(start),
        as.integer(end),as.double(conf),as.integer(composite),
        PACKAGE="cnvpack")
      
      
      CNV.dup = call1$seg
      rm(start,end,cnv.del,cnv.dup,call1) # save space
      
      ##............... END of SUMOVER
      ##............................START of make.region
      ##  ------------------------------------------------------   DELETIONS
      
      REG.del = LOC.del = list()
      lim = LIM[1]
      if(verbose)
        cat('\n DELETION REGIONS: LIMIT = ',lim, 'individuals', '\n')
      
      cnv = CNV.del
      if(composite)
	lim <- quantile(-cnv,probs=high.conf/100)
      
      ## SC: 5 Sept 08
      del = as.numeric(cnv < -lim) ## this is 1 order of magnitude faster!
      ## del = ifelse(cnv < -lim,1,0)
      
      loc = diff(c(0,del,0))
      st = which(loc==1)
      st.chr = chr[st]
      start.pos = pos[st]
      
      en = which(loc== -1) -1
      en.chr = chr[en]
      end.pos = pos[en]
      
      len = end.pos - start.pos +1
      chr.check = (en.chr == st.chr) & (len<10000000)  # same chr, no centromere region
      
      ## collect results
      LOC.del[[1]] = cbind(st,en, st.chr)[chr.check,]
      
      reg = rep(0,length(cnv))  # regions of CNV
      for (ii in 1:length(st))
        {
          reg[st[ii]:en[ii]] = ii * as.numeric(chr.check[ii])
        }
      if(verbose)
        {
          cat('Number of CNV regions across individuals =', sum(chr.check), '\n')
          cat('Average length of CNV regions', mean(len[chr.check]),'\n')
          cat('Median length of CNV regions', median(len[chr.check]),'\n')
          
          len.cut = c(0,1000,10000,50000,1000000)
          cat('Distribution of length','\n')
          print(table(cut(len, len.cut)))
        }
      
      ## collect
      REG.del[[1]] = reg
      
      
      ## Minimum value of each deletion region. Region=0 for background
      REGMIN = tapply(CNV.del,REG.del[[1]], min)[-1]   # remove region 0
      LOC.del[[1]] = cbind(LOC.del[[1]],regmin=REGMIN)
      
      ## get singletons from here:
      freq = table(REGMIN)
      ##print(freq)
      
      
      names(LOC.del) = c(paste('limit=',LIM,sep=""))
      rm(chr.check,del,st,en,len)
      
      
      ##  -------------------------------------------------   DUPLICATIONS
      
      REG.dup = LOC.dup = list()
      lim = LIM[1]
      
      if(verbose)
        cat('\n DUPLICATION REGIONS: LIMIT = ',lim, 'individuals','\n')
      
      cnv = CNV.dup
      if(composite)
	lim <-  quantile(cnv,probs=high.conf/100)
      
      ## SC: 5 Sept
      dup = as.numeric(cnv > lim)
      
      loc = diff(c(0,dup,0))
      st = which(loc==1)
      st.chr = chr[st]
      start.pos = pos[st]
      
      
      en = which(loc== -1) -1
      en.chr = chr[en]
      end.pos = pos[en]
      
      len = end.pos - start.pos +1
      chr.check = (en.chr == st.chr)& (len<10000000)  # same chr, no centromere region
      
      ## collect results
      LOC.dup[[1]] = cbind(st,en, st.chr)[chr.check,]
      
      reg = rep(0,length(cnv))  # regions of CNV
      for (ii in 1:length(st))
        {
          if (chr.check[ii]) reg[st[ii]:en[ii]] = ii
        }
      if(verbose)
        {
          cat('Number of CNV regions across individuals =', sum(chr.check), '\n')
          cat('Average length of CNV regions', mean(len[chr.check]),'\n')
          cat('Median length of CNV regions', median(len[chr.check]),'\n')
          
          len.cut = c(0,1000,10000,50000,1000000)
          cat('Distribution of length','\n')
          print(table(cut(len, len.cut)))
        }
      
      ## collect
      REG.dup[[1]] = reg
      
      
      ## Minimum value of each dupetion region. Region=0 for background
      REGMAX = tapply(CNV.dup,REG.dup[[1]], 'max')[-1]   # remove region 0
      LOC.dup[[1]] = cbind(LOC.dup[[1]],regmax=REGMAX)
      
      ## get singletons from here:
      freq = table(REGMAX)
      ##print(freq)
      
      names(LOC.dup) = c(paste('limit=',LIM,sep=""))
      ##save(REG.del, REG.dup, LOC.del, LOC.dup, file='CNV-regions.RData') #save regions
      rm(chr.check,dup,st,en,len,REG.del,REG.dup)
      
      
      ##............END of make.region
      
      
      ##.................START of reg2indv
      
      ## ........................   DELETIONS
      
      STid = idx[object$startsnp]     
      ENid = idx[object$endsnp]       
      grp.del = grp.dup = list()
      call.del = call.dup = list()
      cnf.del = cnf.dup = list()
      anno.del = anno.dup = list()
      
      samples  = unique(object$sample)
      
      for (j in 1: length(samples))
        {
          sam = samples[j]
          pick = (object$sample %in% sam)  & (object$cn < 2) # deletion
          
          if(sum(pick) > 0) ## If no deletion for subject j skip
            {
              outsam = object[pick,]
              st = STid[pick]   
              en = ENid[pick]   
              
                                        # is the individual region a superset of the common region?        
              st.check = outer(st, LOC.del[[1]][,1], '<=')  
              en.check = outer(en, LOC.del[[1]][,2], '>=')
              check.sp = st.check & en.check
              sp.id = matrix(rep(1:ncol(check.sp),nrow(check.sp)),nrow(check.sp),ncol(check.sp),byrow=TRUE)
              sp.idrow = matrix(rep(1:nrow(check.sp),ncol(check.sp)),nrow(check.sp),ncol(check.sp),byrow=FALSE)
              
                                        #loc1 = c(check.sp %*% c(1:ncol(check.sp)))
              loc1 = c(sp.id[check.sp])
              loc1row = c(sp.idrow[check.sp])
              uloc = unique(loc1)  # unique locations
              
                                        # does the individual region sits inside the common region?        
              st.check = outer(st, LOC.del[[1]][,1], '>=')  
              en.check = outer(en, LOC.del[[1]][,2], '<=')
              check.in = st.check & en.check
              loc2 = c(check.in %*% c(1:ncol(check.in)))
              uloc = c(uloc,unique(loc2))  # unique locations
              
                                        # is the individual region overlap with the common region?        
              st.check = outer(st, LOC.del[[1]][,1], '>=')  
              en.check = outer(en, LOC.del[[1]][,2], '>=')
              over.check = outer(st, LOC.del[[1]][,2], '<=') 
              ## check on shared-length
              share.len = -outer(st,LOC.del[[1]][,2],'-')/c(LOC.del[[1]][,2]-LOC.del[[1]][,1])
              check.ovr = st.check & en.check & over.check & share.len>0 & share.len>=0.8 & share.len<=1
                                        #loc3 = c(check.ovr %*% c(1:ncol(check.ovr)))
              loc3 = c(sp.id[check.ovr])
              loc3row = c(sp.idrow[check.ovr])
              uloc = c(uloc,unique(loc3))  # unique locations
              
              st.check = outer(st, LOC.del[[1]][,1], '<=')  
              en.check = outer(en, LOC.del[[1]][,2], '<=')
              over.check = outer(en, LOC.del[[1]][,1], '>=') 
              ## check on shared-length
              share.len = outer(en,LOC.del[[1]][,1],'-')/c(LOC.del[[1]][,2]-LOC.del[[1]][,1])
              check.ovl = st.check & en.check & over.check & share.len>0 & share.len>=0.8 & share.len<=1
                                        #loc4 = c(check.ovl %*% c(1:ncol(check.ovl)))
              loc4 = c(sp.id[check.ovl])
              loc4row = c(sp.idrow[check.ovl])
              uloc = c(uloc,unique(loc4))  # unique locations
              
              uloc = sort(unique(uloc)) 	
              
              ## median confidence of each region
              if(confidence)
                {
                  cnf1 = tapply(outsam$conf[loc1row],loc1, median)
                  ucnf1 = as.numeric(names(cnf1)) 
                  cnf2 = tapply(outsam$conf,loc2, median)
                  ucnf1 = c(ucnf1,as.numeric(names(cnf2)))
                  cnf3 = tapply(outsam$conf[loc3row],loc3, median)
                  ucnf1 = c(ucnf1,as.numeric(names(cnf3)))
                  cnf4 = tapply(outsam$conf[loc4row],loc4, median)
                  ucnf1 = c(ucnf1,as.numeric(names(cnf4)))
                  cnf = tapply(c(cnf1,cnf2,cnf3,cnf4),ucnf1,median)
                  cnf.del[[j]] = cnf[sort(uloc)>0] # add sort(uloc) May18,2009
                }
              
              
              ## CNV call for each region, if there are multiple calls for one region, choose call with highest conf
              cll1 = tapply(outsam$cn[loc1row],loc1, min)
              ucnf1 = as.numeric(names(cll1)) 
              cll2 = tapply(outsam$cn,loc2, min)
              ucnf1 = c(ucnf1,as.numeric(names(cll2)))
              cll3 = tapply(outsam$cn[loc3row],loc3, min)
              ucnf1 = c(ucnf1,as.numeric(names(cll3)))
              cll4 = tapply(outsam$cn[loc4row],loc4, min)
              ucnf1 = c(ucnf1,as.numeric(names(cll4)))
              
              cll = tapply(c(cll1,cll2,cll3,cll4),ucnf1,min)
              call.del[[j]] = cll[sort(uloc)>0]
              
              
              ## ...... gene annotation
              if(annotated) {
                ## SC: 5 Sept 08: corrected .gene -> \\.gene
                anno1 = tapply(outsam[,grep("\\.gene",colnames(outsam))][loc1row],loc1, unique)
                anno1 = sapply(anno1,paste,collapse=',')
                                        #anno1 = unlist(anqno1)
                ucnf1 = as.numeric(names(anno1)) 
                
                anno2 = tapply(outsam[,grep("\\.gene",colnames(outsam))],loc2, unique)
                anno2 = sapply(anno2,paste,collapse=',')
                                        #anno2 = unlist(anno2)
                ucnf1 = c(ucnf1,as.numeric(names(anno2)))
                
                anno3 = tapply(outsam[,grep("\\.gene",colnames(outsam))][loc3row],loc3, unique)
                anno3 = sapply(anno3,paste,collapse=',')
                                        #anno3 = unlist(anno3)
                ucnf1 = c(ucnf1,as.numeric(names(anno3)))
                
                anno4 = tapply(outsam[,grep("\\.gene",colnames(outsam))][loc4row],loc4, unique)
                anno4 = sapply(anno4,paste,collapse=',')
                                        #anno4 = unlist(anno4)
                ucnf1 = c(ucnf1,as.numeric(names(anno4)))
                
                anno = tapply(c(anno1,anno2,anno3,anno4),ucnf1,unique)
                anno = sapply(anno,paste,collapse=',')
                ucnf1 = as.numeric(names(anno))
                anno.del[[j]] = anno[match(uloc,ucnf1)][uloc>0]
              }
              
              ## ...... individual j's regions
              grp.del[[j]] = sort(uloc[uloc>0])
              
            }

          ## ........................ DUPLICATIONS
          pick = (object$sample == sam)  & (object$cn > 2)   # duplications

          if(sum(pick) > 0)
            {
              outsam = object[pick,]  
              st = STid[pick]  
              en = ENid[pick]   
              
              
                                        # is the individual region a superset of the common region?        
              st.check = outer(st, LOC.dup[[1]][,1], '<=')  
              en.check = outer(en, LOC.dup[[1]][,2], '>=')
              check.sp = st.check & en.check
              sp.id = matrix(rep(1:ncol(check.sp),nrow(check.sp)),nrow(check.sp),ncol(check.sp),byrow=TRUE)
              sp.idrow = matrix(rep(1:nrow(check.sp),ncol(check.sp)),nrow(check.sp),ncol(check.sp),byrow=FALSE)
              
                                        #loc1 = c(check.sp %*% c(1:ncol(check.sp)))
              loc1 = c(sp.id[check.sp])
              loc1row = c(sp.idrow[check.sp])
              uloc = unique(loc1)  # unique locations
              
                                        # does the individual region sits inside the common region?        
              st.check = outer(st, LOC.dup[[1]][,1], '>=')  
              en.check = outer(en, LOC.dup[[1]][,2], '<=')
              check.in = st.check & en.check
              loc2 = c(check.in %*% c(1:ncol(check.in)))
              uloc = c(uloc,unique(loc2))  # unique locations
              
                                        # is the individual region overlap with the common region?        
              st.check = outer(st, LOC.dup[[1]][,1], '>=')  
              en.check = outer(en, LOC.dup[[1]][,2], '>=')
              over.check = outer(st, LOC.dup[[1]][,2], '<=') 
              ## check on shared-length
              share.len = -outer(st,LOC.dup[[1]][,2],'-')/c(LOC.dup[[1]][,2]-LOC.dup[[1]][,1])
              check.ovr = st.check & en.check & over.check & share.len>0 & share.len>=0.8 & share.len<=1
                                        #loc3 = c(check.ovr %*% c(1:ncol(check.ovr)))
              loc3 = c(sp.id[check.ovr])
              loc3row = c(sp.idrow[check.ovr])
              uloc = c(uloc,unique(loc3))  # unique locations
              
              st.check = outer(st, LOC.dup[[1]][,1], '<=')  
              en.check = outer(en, LOC.dup[[1]][,2], '<=')
              over.check = outer(en, LOC.dup[[1]][,1], '>=') 
              ## check on shared-length
              share.len = outer(en,LOC.dup[[1]][,1],'-')/c(LOC.dup[[1]][,2]-LOC.dup[[1]][,1])
              check.ovl = st.check & en.check & over.check & share.len>0 & share.len>=0.8 & share.len<=1
                                        #loc4 = c(check.ovl %*% c(1:ncol(check.ovl)))
              loc4 = c(sp.id[check.ovl])
              loc4row = c(sp.idrow[check.ovl])
              uloc = c(uloc,unique(loc4))  # unique locations
              
              uloc = sort(unique(uloc)) 	
              
              ## median confidence of each region
              if(confidence)
                {
                  cnf1 = tapply(outsam$conf[loc1row],loc1, median)
                  ucnf1 = as.numeric(names(cnf1)) 
                  cnf2 = tapply(outsam$conf,loc2, median)
                  ucnf1 = c(ucnf1,as.numeric(names(cnf2)))
                  cnf3 = tapply(outsam$conf[loc3row],loc3, median)
                  ucnf1 = c(ucnf1,as.numeric(names(cnf3)))
                  cnf4 = tapply(outsam$conf[loc4row],loc4, median)
                  ucnf1 = c(ucnf1,as.numeric(names(cnf4)))
                  cnf = tapply(c(cnf1,cnf2,cnf3,cnf4),ucnf1,median)
                  cnf.dup[[j]] = cnf[sort(uloc)>0] # add sort(uloc) May18,2009
                }
              
              
              ## CNV call for each region, if there are multiple calla for one region, choose call with highest conf
              cll1 = tapply(outsam$cn[loc1row],loc1, min)
              ucnf1 = as.numeric(names(cll1)) 
              cll2 = tapply(outsam$cn,loc2, min)
              ucnf1 = c(ucnf1,as.numeric(names(cll2)))
              cll3 = tapply(outsam$cn[loc3row],loc3, min)
              ucnf1 = c(ucnf1,as.numeric(names(cll3)))
              cll4 = tapply(outsam$cn[loc4row],loc4, min)
              ucnf1 = c(ucnf1,as.numeric(names(cll4)))
              
              cll = tapply(c(cll1,cll2,cll3,cll4),ucnf1,min)
              call.dup[[j]] = cll[sort(uloc)>0]
              
              
              ## ...... gene annotation
              if(annotated) {
                ## SC: 5 Sept 08: corrected .gene -> \\.gene
                anno1 = tapply(outsam[,grep("\\.gene",colnames(outsam))][loc1row],loc1, unique)
                anno1 = sapply(anno1,paste,collapse=',')
                                        #anno1 = unlist(anno1)
                ucnf1 = as.numeric(names(anno1)) 
                
                anno2 = tapply(outsam[,grep("\\.gene",colnames(outsam))],loc2, unique)
                anno2 = sapply(anno2,paste,collapse=',')
                                        #anno2 = unlist(anno2)
                ucnf1 = c(ucnf1,as.numeric(names(anno2)))
                
                anno3 = tapply(outsam[,grep("\\.gene",colnames(outsam))][loc3row],loc3, unique)
                anno3 = sapply(anno3,paste,collapse=',')
                                        #anno3 = unlist(anno3)
                ucnf1 = c(ucnf1,as.numeric(names(anno3)))
                
                anno4 = tapply(outsam[,grep("\\.gene",colnames(outsam))][loc4row],loc4, unique)
                anno4 = sapply(anno4,paste,collapse=',')
                                        #anno4 = unlist(anno4)
                ucnf1 = c(ucnf1,as.numeric(names(anno4)))
                
                anno = tapply(c(anno1,anno2,anno3,anno4),ucnf1,unique)
                anno = sapply(anno,paste,collapse=',')
                ucnf1 = as.numeric(names(anno))
                anno.dup[[j]] = anno[match(uloc,ucnf1)][uloc>0]
              }
              
              ## ...... individual j's regions
              grp.dup[[j]] = sort(uloc[uloc>0])
            }
        }
      
      INDIV.del = INDIV.dup = list()
      
      INDIV.del[[1]] = grp.del
      INDIV.dup[[1]] = grp.dup
      
      INDIV.del[[2]] = cnf.del
      INDIV.dup[[2]] = cnf.dup
      
      INDIV.del[[3]] = call.del
      INDIV.dup[[3]] = call.dup
      
      INDIV.del[[4]] = anno.del
      INDIV.dup[[4]] = anno.dup
      
      INDIV.del[[5]] = LOC.del
      INDIV.dup[[5]] = LOC.dup
      
      names(INDIV.del) = c("delID","conf","call","anno","loc")
      names(INDIV.dup) = c("dupID","conf","call","anno","loc")
      
      INDIV.cnv = list()
      INDIV.cnv[[1]] = INDIV.del
      INDIV.cnv[[2]] = INDIV.dup
      names(INDIV.cnv) = c("del","dup")
      
    if(!missing(pheno.data))
      {
        INDIV.cnv[[3]] = pheno.data
        names(INDIV.cnv) <- c("del","dup","pheno")
      }
   } # end of cnv.abnormality = 'separate'


 ## ................................. START of cnv.abnormality=='both'

  if(cnv.abnormality=="both") {
    pick = object$cn != 2  #deletion+dups
    start = idx[object$startsnp[pick]]
    end =  idx[object$endsnp[pick]]
    nsubj = length(unique(object$sample))    
    conf = object$conf[pick]
  
    n = length(cnv.del)
    m = length(start)
    
    call1 = .Fortran("list2seg", 
      as.integer(n),
      seg = as.double(cnv.del),
      as.integer(m),
      as.integer(start),
      as.integer(end),as.double(conf),
      as.integer(composite),PACKAGE="cnvpack")
    
    CNV.del = -call1$seg

 ##  ------------------------------------------------------   REGIONS

    REG.del = LOC.del = list()
    lim = LIM[1]
    if(verbose)
      cat('\n CNV REGIONS: LIMIT = ',lim, 'individuals', '\n')
    
    cnv = CNV.del
    if(composite) 
	lim <- quantile(-cnv,probs=high.conf/100)
    
    ## SC: 5 Sept 08
    del = as.numeric(cnv < -lim) ## this is 1 order of magnitude faster!
    ## del = ifelse(cnv < -lim,1,0)

    loc = diff(c(0,del,0))

    st = which(loc==1)
    st.chr = chr[st]
    start.pos = pos[st]
    
    en = which(loc== -1) -1
    en.chr = chr[en]
    end.pos = pos[en]

    len = end.pos - start.pos +1
    chr.check = (en.chr == st.chr) & (len<10000000)  # same chr, no centromere region
    

    nocluster = cluster.method=='none'

    if(!nocluster){		## start of clustering
    start1 = start.pos[chr.check]
    end1 = end.pos[chr.check]
    chr_region1 = st.chr[chr.check]
    CNP_start = CNP_end = g = st2 = en2 = st.chr2 = len2 = sample = conf = cn = gene = plotst.del = ploten.del = list()
    temp = object

    temp$chr = as.numeric(substring(object$chr,4,5)) #get chr num
    
    for(m in 1:length(start1)){ #i corresponds to CNP ID
        
        start = start1[m]
        end = end1[m]
        chr_region = as.numeric(chr_region1[m])
   
        #extract the individual CNVs that overlap with this CNP region
        #check super set
        check1 = temp$start<=start
        check2 = temp$start>=end
        checkchr = temp$chr==chr_region
        superset = check1 & check2 & checkchr
        
        #check inside
        check1 = temp$start>=start
        check2 = temp$end<=end
        inside = check1 & check2 & checkchr
        
        #check 80% overlap
        check1 = temp$start>=start & temp$end>=end
        check1over = ((end-temp$start+1)/(end-start+1))>=0.8
        check2 = temp$start<=start & temp$end<=end
        check2over = ((temp$end-start+1)/(end-start+1))>=0.8 
        overlap1 = check1 & check1over & checkchr
        overlap2 = check2 & check2over & checkchr 
        
        indiv = temp[(superset|inside|overlap1|overlap2),]
        samples  = unique(temp$sample)
        sample[[m]] = match(indiv$sample,samples)
        conf[[m]] = indiv$conf
        cn[[m]] = indiv$cn
        #num is the number of individual CNV that overlaps with this CNP region
        num = length(indiv$start)
        if(num>1){      #we can only do clustering when the number of individual CNVs is greater than 1
            # make sure individual regions do not go beyond CNP region
            for(k in 1:num){
                indiv$start[k] = max(start,indiv$start[k])
                indiv$end[k] = min(end,indiv$end[k])
            }
    
            #get disimilarity matrix using Jaccard's method       b+c/(a+b+c)
            dissim = matrix(0,nrow=num,ncol=num)
    
            for(i in 1:(num-1)){
                for(j in (i+1):num){
                    dissim[j,i] = min((abs(indiv$start[i]-indiv$start[j])+abs(indiv$end[i]-indiv$end[j]))/(max(indiv$end[i],indiv$end[j])-min(indiv$start[i],indiv$start[j])+1),1)  
                }
            }
             d = as.dist(dissim, diag = FALSE, upper = FALSE)
            hc1 = hclust(d, method = cluster.method, members=NULL)
            groups = cutree(hc1, h=cluster.limit) 
            g[[m]] = groups
            CNP_start[[m]] = min(indiv$start[groups==1])
            CNP_end[[m]] = max(indiv$end[groups==1])
    
            #define new start and end for the sub regions
            if(sum(I(groups==2))!=0){ #means there is more than 1 cluster
                for(n in 2:max(groups)){
      
                    CNP_start[[m]][n]=min(indiv$start[groups==n])
                    CNP_end[[m]][n]=max(indiv$end[groups==n])
            
                }
                #names(CNP_start[[m]])=paste(m,".",1:max(groups))
                names(CNP_start[[m]])= m + (1:max(groups))/100
                #names(sample[[m]])=paste(m,".",g[[m]])
                names(sample[[m]])= m + g[[m]]/100
            }   
            if(sum(I(groups==2))==0){ #means all individuals in same cluster
                
                names(CNP_start[[m]])=m
                names(sample[[m]])=rep(m,length(sample[[m]]))
            } 
            

      } 
    if(num==1){  #when there is only 1 indiv, then use original region
      CNP_start[[m]] = indiv$start
      CNP_end[[m]] =  indiv$end
      g[[m]] = 1
      names(CNP_start[[m]])=m
      names(sample[[m]])=m
    }    
   st2[[m]]=match(CNP_start[[m]],annotation$Position)
   en2[[m]]=match(CNP_end[[m]],annotation$Position)
   st.chr2[[m]]=rep(chr_region,max(g[[m]])) 
   len2[[m]] = CNP_end[[m]] - CNP_start[[m]] + 1
   plotst.del[[m]] = indiv$start
   ploten.del[[m]] = indiv$end
   gene[[m]] = indiv[,grep("\\.gene",colnames(indiv))]
   }
   ##SM collect results
   LOC.del[[1]] = cbind(unlist(st2),unlist(en2),unlist(st.chr2))
   names(LOC.del) = c(paste('limit=',LIM,sep=""))
   colnames(LOC.del[[1]])=c("st","en","chr")
   rownames(LOC.del[[1]])=names(unlist(CNP_start))
   if(verbose){
        cat('Number of CNV regions across individuals =', length(unlist(st2)), '\n')
        cat('Average length of CNV regions', mean(unlist(len2)),'\n')
        cat('Median length of CNV regions', median(unlist(len2)),'\n')
        len.cut = c(0,1000,10000,50000,1000000)
        cat('Distribution of length','\n')
        print(table(cut(unlist(len2), len.cut)))
   }
     
    ## .......................................   REGIONS AFTER CLUSTERING

    reg = list()
    regconf = regcall = reganno = list()
    samples  = unique(object$sample)
    for (j in 1: length(samples)){
          final_id=NULL
          for(i in 1:length(CNP_start)){   ##number of big regions
              regID = names(sample[[i]][which(j==sample[[i]])])    

              if(length(regID)>0){
                  unqid = unique(regID)
                  final_id = c(final_id,unqid)
              }
         }
	 if(length(final_id)>0)
            reg[[j]]=noquote(final_id[!is.na(final_id)])
	 if(length(final_id)==0)
	    reg[[j]] = NULL

     
         
        ## choose call with highest confidence
          if(confidence) {
        
          final_conf=NULL
          final_call=NULL
          
          for(i in 1:length(CNP_start)){   ##number of big regions
            
              cnf = conf[[i]][which(j==sample[[i]])]
              call = cn[[i]][which(j==sample[[i]])]
              regID = names(sample[[i]][which(j==sample[[i]])])
              
              if(length(regID)>0){
                   unqid = unique(regID)
                   bestcall=NULL
                   highconf=NULL
              
                   for(ii in 1:length(unqid)){
                        maxconf = max(cnf[which(regID==unqid[ii])])
  	                    bestcall = c(bestcall,call[which(cnf==maxconf &regID==unqid[ii])])
	                      highconf = c(highconf,cnf[which(cnf==maxconf &regID==unqid[ii])])
	
                  }     
                
                  final_conf = c(final_conf,highconf)
                  final_call = c(final_call,bestcall)
            }
         
           }
           regconf[[j]]=final_conf[!is.na(final_conf)]
           regcall[[j]]=final_call[!is.na(final_call)]
           if(length(final_id)>0) {
              names(regconf[[j]])=reg[[j]]
              names(regcall[[j]])=reg[[j]]
	   }
          }   #end if confidence
                   
          if(!confidence){
              final_call=NULL
              for(i in 1:length(CNP_start)){   ##number of big regions
                  regID = names(sample[[i]][which(j==sample[[i]])])
                  call = cn[[i]][which(j==sample[[i]])]
                  if(length(regID)>0){
                      unqid = unique(regID)
                      bestcall=NULL
                      for(ii in 1:length(unqid)){
                           bestcall = c(bestcall,min(call[which(regID==unqid[ii])]))
	                    }     
                      final_call = c(final_call,bestcall)
                  }
             }
             regcall[[j]]=final_call[!is.na(final_call)]
             if(length(final_id)>0) 
               names(regcall[[j]])=reg[[j]]
          } #end if(!confidence) 
          
          if(annotated){
              final_anno=NULL
          
              for(i in 1:length(CNP_start)){   ##number of big regions
            
              regID = names(sample[[i]][which(j==sample[[i]])])
              anno = gene[[i]][which(j==sample[[i]])]
              if(length(regID)>0){
                   unqid = unique(regID)
                   bestanno=NULL
                                
                   for(ii in 1:length(unqid)){
                        bestanno = c(bestanno,anno[which(regID==unqid[ii])])
	
                  }     
                
                  final_anno = c(final_anno,bestanno)
                
            }
         
           }
           reganno[[j]]=final_anno[!is.na(final_anno)]
           if(length(final_id)>0) 
            names(reganno[[j]])=reg[[j]]
          }
      }
      grp.del = cnf.del = call.del = anno.del = INDIV.del = INDIV.cnv = list()
      grp.del=reg
      cnf.del=regconf
      call.del=regcall
      anno.del=reganno
      clustergroups.del=g
    } #end of cluster 


   if(nocluster) {
    ## collect results
    LOC.del[[1]] = cbind(st,en, st.chr)[chr.check,]
    
    reg = rep(0,length(cnv))  # regions of CNV
    for (ii in 1:length(st))
      {
        reg[st[ii]:en[ii]] = ii * as.numeric(chr.check[ii])
      }
    if(verbose)
      {
        cat('Number of CNV regions across individuals =', sum(chr.check), '\n')
        cat('Average length of CNV regions', mean(len[chr.check]),'\n')
        cat('Median length of CNV regions', median(len[chr.check]),'\n')
        
        len.cut = c(0,1000,10000,50000,1000000)
        cat('Distribution of length','\n')
        print(table(cut(len, len.cut)))
      }
    
    ## collect
    REG.del[[1]] = reg
    
    
    ## Minimum value of each region. Region=0 for background
    REGMIN = tapply(CNV.del,REG.del[[1]], 'min')[-1]   # remove region 0
    LOC.del[[1]] = cbind(LOC.del[[1]],regmin=REGMIN)
    
    ## get singletons from here:
    freq = table(REGMIN)
    ##print(freq)
    names(LOC.del) = c(paste('limit=',LIM,sep=""))
    rm(chr.check,del,st,en,len)

    ##.................START of reg2indv


    ## .......................................   REGIONS

    STid = idx[object$startsnp]     
    ENid = idx[object$endsnp]       
    grp.del = list()
    cnf.del = call.del = list()
    anno.del = list()
    
    samples  = unique(object$sample)

    for (j in 1: length(samples))
      {
        sam = samples[j]
        pick = (object$sample == sam)  
        outsam = object[pick,]  
        st = STid[pick]   
        en = ENid[pick]   

	# is the individual region a superset of the common region?        
        st.check = outer(st, LOC.del[[1]][,1], '<=')  
        en.check = outer(en, LOC.del[[1]][,2], '>=')
        check.sp = st.check & en.check
        sp.id = matrix(rep(1:ncol(check.sp),nrow(check.sp)),nrow(check.sp),ncol(check.sp),byrow=TRUE)
        sp.idrow = matrix(rep(1:nrow(check.sp),ncol(check.sp)),nrow(check.sp),ncol(check.sp),byrow=FALSE)

        #loc1 = c(check.sp %*% c(1:ncol(check.sp)))
        loc1 = c(sp.id[check.sp])
        loc1row = c(sp.idrow[check.sp])
        uloc = unique(loc1)  # unique locations

	# does the individual region sits inside the common region?        
        st.check = outer(st, LOC.del[[1]][,1], '>=')  
        en.check = outer(en, LOC.del[[1]][,2], '<=')
        check.in = st.check & en.check
        loc2 = c(check.in %*% c(1:ncol(check.in)))
        uloc = c(uloc,unique(loc2))  # unique locations

	# does the individual region overlap with the common region?        
        st.check = outer(st, LOC.del[[1]][,1], '>=')  
        en.check = outer(en, LOC.del[[1]][,2], '>=')
        over.check = outer(st, LOC.del[[1]][,2], '<=') 
        ## check on shared-length
        share.len = -outer(st,LOC.del[[1]][,2],'-')/c(LOC.del[[1]][,2]-LOC.del[[1]][,1])
        check.ovr = st.check & en.check & over.check & share.len>0 & share.len>=0.8 & share.len<=1
        #loc3 = c(check.ovr %*% c(1:ncol(check.ovr)))
	  loc3 = c(sp.id[check.ovr])
        loc3row = c(sp.idrow[check.ovr])
        uloc = c(uloc,unique(loc3))  # unique locations

        st.check = outer(st, LOC.del[[1]][,1], '<=')  
        en.check = outer(en, LOC.del[[1]][,2], '<=')
        over.check = outer(en, LOC.del[[1]][,1], '>=') 
        ## check on shared-length
        share.len = outer(en,LOC.del[[1]][,1],'-')/c(LOC.del[[1]][,2]-LOC.del[[1]][,1])
        check.ovl = st.check & en.check & over.check & share.len>0 & share.len>=0.8 & share.len<=1
        #loc4 = c(check.ovl %*% c(1:ncol(check.ovl)))
	  loc4 = c(sp.id[check.ovl])
        loc4row = c(sp.idrow[check.ovl])
        uloc = c(uloc,unique(loc4))  # unique locations

	  uloc = sort(unique(uloc)) 	
        
        ## median confidence of each region
        if(confidence)
          {
            cnf1 = tapply(outsam$conf[loc1row],loc1, median)
		ucnf1 = as.numeric(names(cnf1)) 
            cnf2 = tapply(outsam$conf,loc2, median)
		ucnf1 = c(ucnf1,as.numeric(names(cnf2)))
            cnf3 = tapply(outsam$conf[loc3row],loc3, median)
		ucnf1 = c(ucnf1,as.numeric(names(cnf3)))
            cnf4 = tapply(outsam$conf[loc4row],loc4, median)
		ucnf1 = c(ucnf1,as.numeric(names(cnf4)))
 		cnf = tapply(c(cnf1,cnf2,cnf3,cnf4),ucnf1,median)
            cnf.del[[j]] = cnf[sort(uloc)>0] # add sort(uloc) May18,2009
          }


     ## CNV call for each region, if there are multiple calla for one region, choose call with highest conf
        cll1 = tapply(outsam$cn[loc1row],loc1, min)
	  ucnf1 = as.numeric(names(cll1)) 
        cll2 = tapply(outsam$cn,loc2, min)
	  ucnf1 = c(ucnf1,as.numeric(names(cll2)))
        cll3 = tapply(outsam$cn[loc3row],loc3, min)
	  ucnf1 = c(ucnf1,as.numeric(names(cll3)))
        cll4 = tapply(outsam$cn[loc4row],loc4, min)
	  ucnf1 = c(ucnf1,as.numeric(names(cll4)))

        cll = tapply(c(cll1,cll2,cll3,cll4),ucnf1,min)
        call.del[[j]] = cll[sort(uloc)>0]
        
      
        ## ...... gene annotation
     	if(annotated) {
          ## SC: 5 Sept 08: corrected .gene -> \\.gene
          anno1 = tapply(outsam[,grep("\\.gene",colnames(outsam))][loc1row],loc1, unique)
          anno1 = sapply(anno1,paste,collapse=',')
          #anno1 = unlist(anno1)
	    ucnf1 = as.numeric(names(anno1)) 

	    anno2 = tapply(outsam[,grep("\\.gene",colnames(outsam))],loc2, unique)
          anno2 = sapply(anno2,paste,collapse=',')
          #anno2 = unlist(anno2)
	    ucnf1 = c(ucnf1,as.numeric(names(anno2)))

          anno3 = tapply(outsam[,grep("\\.gene",colnames(outsam))][loc3row],loc3, unique)
          anno3 = sapply(anno3,paste,collapse=',')
          #anno3 = unlist(anno3)
	    ucnf1 = c(ucnf1,as.numeric(names(anno3)))

          anno4 = tapply(outsam[,grep("\\.gene",colnames(outsam))][loc4row],loc4, unique)
          anno4 = sapply(anno4,paste,collapse=',')
          #anno4 = unlist(anno4)
	    ucnf1 = c(ucnf1,as.numeric(names(anno4)))

          anno = tapply(c(anno1,anno2,anno3,anno4),ucnf1,unique)
          anno = sapply(anno,paste,collapse=',')
	    ucnf1 = as.numeric(names(anno))
          anno.del[[j]] = anno[match(uloc,ucnf1)][uloc>0]
        }

        ## ...... individual j's regions
        grp.del[[j]] = sort(uloc[uloc>0])

    }
   } # end of nocluster

    INDIV.del = list()
    
    INDIV.del[[1]] = grp.del
    INDIV.del[[2]] = cnf.del
    INDIV.del[[3]] = call.del
    INDIV.del[[4]] = anno.del
    INDIV.del[[5]] = LOC.del

    if(!nocluster) {
      INDIV.del[[6]] = plotst.del
      INDIV.del[[7]] = ploten.del
      INDIV.del[[8]] = clustergroups.del
      names(INDIV.del) = c("regID","conf","call","anno","loc","plotst","ploten","clustergroups")
    } 

    if(nocluster)
      names(INDIV.del) = c("regID","conf","call","anno","loc")
    
    INDIV.cnv = list()
    INDIV.cnv[[1]] = INDIV.del
    names(INDIV.cnv) = c("reg")
    
    if(!missing(pheno.data))
      {
        INDIV.cnv[[2]] = pheno.data
        names(INDIV.cnv) <- c("reg","pheno")
      }

   } # end of cnv.abnormality='both'

    class(INDIV.cnv) = c("cnv","list")
    attr(INDIV.cnv,"both") <- (cnv.abnormality=='both')
    attr(INDIV.cnv,"annotation") <- deparse(substitute(annotation))
    
    return(INDIV.cnv)
       
  } # end of function

